/*********************************************************
File:   ads.c
Author: Jan Rohde
Date:   4.6.07
Description:  mainfile for the Attitude Determination System.
              After running the initialization programs it
              goes into a continues while-loop.
*********************************************************/


#include <__cross_studio_io.h>
#include "ads.h"


void ads(void){
  /*
    creating variables
  */
  int state = IDLE;
  float **J, **M;
  float **K, **P, **H, **G, **R, **Q, **Kw, **Ke, **P_bar, **P_bar_next, **I, **kxyz;
  float **Phi_r, **Phi_tot, **F_ATT, **F_VEL;
  float **eps_skew, **temp_eps_skew;
  
  float **y_ref, **y_meas, **x_bar_next, **X_hat, **w_bar_next, **w_bar, **w_est, **x_est;
  Quaternion q_hat, q_bar, deps, Deps, delta_eps_w, eta, q_est;
  Quaternion q_hat_next = {0.5f, {0.5f, 0.5f, 0.5f}};
  Quaternion q_bar_next = {0.8289f, {-0.1881f, 0.4922f, -0.1881f}};
  Quaternion q_error = {0.0f, {0.0f, 0.0f, 0.0f}};
  Quaternion vQuaternions[10];

  float **vPointers[28];
  vPointers[0] = J;
  vPointers[1] = M;
  vPointers[2] = P;
  vPointers[3] = K;
  vPointers[4] = H;
  vPointers[5] = R;
  vPointers[6] = G;
  vPointers[7] = Q;
  vPointers[8] = Ke;
  vPointers[9] = Kw;
  vPointers[10] = Phi_r;
  vPointers[11] = Phi_tot;
  vPointers[12] = P_bar;
  vPointers[12] = P_bar_next;
  vPointers[14] = y_meas;
  vPointers[15] = y_ref;
  vPointers[16] = w_bar_next;
  vPointers[17] = x_bar_next;
  vPointers[18] = w_bar;
  vPointers[19] = w_est;
  vPointers[20] = x_est;
  vPointers[21] = X_hat;
  vPointers[22] = eps_skew;
  vPointers[23] = temp_eps_skew;
  vPointers[24] = F_ATT;
  vPointers[25] = F_VEL;
  vPointers[26] = kxyz;
  vPointers[27] = I;

  vQuaternions[0] = q_hat;
  vQuaternions[1] = q_hat_next;
  vQuaternions[2] = q_bar_next;
  vQuaternions[3] = q_bar;
  vQuaternions[4] = deps;
  vQuaternions[5] = Deps;
  vQuaternions[6] = delta_eps_w;
  vQuaternions[7] = eta;
  vQuaternions[8] = q_est;
  vQuaternions[9] = q_error;

  /*
    initializing uc
  */
  init_uc();
  
  while(1){

    switch(state){
      case IDLE:
        //Upon receiving a START signal the state is changed to INITIATE_SYSTEM
        break;

      case INITIATE_SYSTEM:
        /*
          running init files 
        */
        initEKF(vPointers, vQuaternions[1]);
        sun_sensor_init();
        magnetometer_communication_init();
        twi_communication_init();

        break;

      case ESTIMATE:
        /*
          The estimator is started and new estimated states produced
        */
        /*
          Get Measurements
        */
        GetMeasurement(y_meas);                 //creates a new set of measurements
        /*
          GAUSS-NEWTON ALGORITHM, creating Jacobie-matrix and estimate new optimal quaternion, q_hat_next
        */
        GenerateJacobian(vPointers, vQuaternions);
        CreateM(vPointers, vQuaternions);
        Calculate_q_hat(vPointers, vQuaternions);
        /*
          Computing Kalman Gain K
        */
        CreateK(vPointers, vQuaternions);
        /*
          Updating estimation error
        */
        UpdateEstimationError(vPointers, vQuaternions);
        /*
          Updating Error Covariance P (6 x 6)
        */
        UpdateErrorCovariance(vPointers, vQuaternions);
        /*
          Forward Euler Integration: Finding Phi_r and Phi_tot
          Phi_r = eye(6) + linearsystemR(x_est);
          Phi_tot = nonLinearProp(x_est);
          x_est = [eta_hat eps_1_hat eps_2_hat eps_3_hat w_est_x w_est_y w_est_z]
        */
        UpdateStateTransitionMatrices(vPointers, vQuaternions);
        /*
          state propagation
        */
        StatePropagation(vPointers, vQuaternions);
        /*
          Transmit estimates
        */
        TransmitEstimates(X_hat);
        break;

      case COMMUNICATE:
        /*
          This state is used for transmitting and receiving messages on the twi-bus
          Uses the twi functions from twi.c
        */
        break;
    }
  }
}

/***********************************************************************
** Initialization functions
***********************************************************************/
void initEKF(float **vMatrix[], Quaternion q_bar_init){
  float kx, ky, kz, Ix, Iy, Iz;
  /*
  initializing the matrices used directly in the EKF
  */
  vMatrix[0] = matrix(1, 6, 1, 4);
  matrix_zeros(vMatrix[0], 6, 4);              //(6 x 4) Jacobien matrix (J)
  
  vMatrix[1] = matrix(1, NX, 1, NX);
  matrix_zeros(vMatrix[1], NX, NX);            //(6 x 6) Rotation matrix (M)

  vMatrix[2] = matrix(1, NX, 1, NX);
  matrix_zeros(vMatrix[2], NX, NX);           //(6 x 6) error covariance matrix (P)
  
  vMatrix[3] = matrix(1, NX, 1, 3);
  matrix_zeros(vMatrix[3], NX, 3);             //(6 x 3) Kalman gain matrix (K)
  
  vMatrix[4] = matrix(1, 3, 1, NY);
  matrix_zeros(vMatrix[4], 3, NY);             //(3 x 6) measurement matrix (H)
  
  vMatrix[5] = matrix(1, 3, 1, 3);
  matrix_diag(vMatrix[5], R_dv, 3);            //(3 x 3) covariance matrix, meas_noise (R)

  vMatrix[6] = matrix(1, NX, 1, NX);
  matrix_ident(vMatrix[6], NX);                //(6 x 6) matrix (G)
  
  vMatrix[7] = matrix(1, NX, 1, NX);
  matrix_diag(vMatrix[7], Q_dv, NX);           //(6 x 6) covariance matrix, process_noise (Q)
  
  vMatrix[8] = matrix(1, 3, 1, 3);
  matrix_zeros(vMatrix[8], 3, 3);              //(3 x 3) upper half of Kalman gain matrix (Ke)
  
  vMatrix[9] = matrix(1, 3, 1, 3);
  matrix_zeros(vMatrix[9], 3, 3);              //(3 x 3) lower half of Kalman gain matrix (Kw)

  vMatrix[10] = matrix(1, NX, 1, NX);
  matrix_zeros(vMatrix[10], NX, NX);               //(6 x 6) Reduced system Matrix (Phi_r)

  vMatrix[11] = matrix(1, NX, 1, 1);
  matrix_zeros(vMatrix[11], NX, 1);                //(6 x 1) Total system (Phi_tot)
  
  vMatrix[12] = matrix(1, NX, 1, NX);
  matrix_zeros(vMatrix[12], NX, NX);           //(6 x 6) error covariance matrix (P_bar)

  vMatrix[13] = matrix(1, NX, 1, NX);
  matrix_diag(vMatrix[13], P_dv, NX);           //(6 x 6) error covariance matrix (P_bar_next)

  /*
  initializing the vectors used directly in the EKF
  */
  vMatrix[14] = matrix(1, 6, 1, 1);
  matrix_zeros(vMatrix[14], 6, 1);                  //(6 x 1) measurement vector (y_meas)
                                                    //y_meas = [bx by bz sx sy sz]
  vMatrix[15] = matrix(1, 6, 1, 1);
  matrix_zeros(vMatrix[15], 6, 1);                  //(6 x 1) reference vector (y_ref)
  
  vMatrix[16] = matrix(1, 3, 1, 1);
  matrix_zeros(vMatrix[16], 3, 1);                  //(3 x 1) omega_bar (w_bar_next)
  vMatrix[16][1][1] = 0.001;
  vMatrix[16][2][1] = 0.001;
  vMatrix[16][3][1] = 0.001;

  vMatrix[17] = matrix(1, 7, 1, 1);
  matrix_zeros(vMatrix[17], 7, 1);                  //(7 x 1) estimated states (x_bar_next)
  vMatrix[17][1][1] = q_bar_init.eta;
  vMatrix[17][2][1] = q_bar_init.eps[0];
  vMatrix[17][3][1] = q_bar_init.eps[1];
  vMatrix[17][4][1] = q_bar_init.eps[2];
  vMatrix[17][5][1] = vMatrix[14][1][1];
  vMatrix[17][6][1] = vMatrix[14][2][1];
  vMatrix[17][7][1] = vMatrix[14][3][1];

  vMatrix[18] = matrix(1, 3, 1, 1);
  matrix_zeros(vMatrix[18], 3, 1);           //(6 x 1) error covariance matrix (w_bar)

  vMatrix[19] = matrix(1, 3, 1, 1);
  matrix_zeros(vMatrix[19], 3, 1);           //(6 x 1) error covariance matrix (w_est)

  vMatrix[20] = matrix(1, 7, 1, 1);
  matrix_zeros(vMatrix[20], 7, 1);           //(6 x 1) error covariance matrix (x_est)

  vMatrix[21] = matrix(1, 7, 1, 1);
  matrix_zeros(vMatrix[21], 7, 1);                  //(7 x 1) (X_hat)
    
  /*
  initializing various other matrices
  */
  vMatrix[22] = matrix(1, 3, 1, 3);
  matrix_zeros(vMatrix[22], 3, 3);                  // (3 x 3) skew symetric matrix of epsilon
  
  vMatrix[23] = matrix(1, 3, 1, 3);                 //(3 x 3) temp_skew_matrix
  matrix_zeros(vMatrix[23], 3, 3);

  vMatrix[24] = matrix(1, 3, 1, 6);

  vMatrix[25] = matrix(1, 3, 1, 6);

  //********************************************************************************
  vMatrix[26] = matrix(1, 3, 1, 1);                  // [Kx Ky Kz]^T
  vMatrix[27] = matrix(1, 3, 1, 3);                  // I - inertia matrix

  Ix = 0.0061f;
  Iy = 0.0061f;
  Iz = 0.0028f;

  kx = (Iy-Iz)/Ix;
  ky = (Ix-Iz)/Iy;
  kz = (Ix-Iy)/Iz;

  vMatrix[26][1][1] = kx;
  vMatrix[26][2][1] = ky;
  vMatrix[26][3][1] = kz;

  matrix_ident(vMatrix[27], 3);
  vMatrix[27][1][1] = vMatrix[27][1][1] * kx;
  vMatrix[27][2][2] = vMatrix[27][2][2] * ky;
  vMatrix[27][3][3] = vMatrix[27][3][3] * kz;
}

void init_uc(void){
  volatile unsigned int i;
  /*
    Initializing the controller
  */
  //initiating Xtal-2
  WDTCTL = WDTPW + WDTHOLD;
  P5DIR |= 0x10;
  P5SEL |= 0x10;
  BCSCTL1 &= ~XT2OFF;

  do{
    IFG1 &= ~OFIFG;
    for (i = 0xFF; i > 0; i--);
  }
  while((IFG1 & OFIFG));

  BCSCTL2 |= SELM_2;
  //Xtal-2 active

}

void twi_communication_init(void){

}

void sun_sensor_init(void){
                            //TimerB is used to measure the frequency
                            //of the light-to-frequency diodes
  int i;
  WDTCTL =  WDTPW + WDTHOLD; 
  BCSCTL1 |= XTS;           //using 8MHz X-tal

  do{
    IFG1 &= ~OFIFG;
    for (i = 0xFF; i > 0; i--);
  }
  while((IFG1 & OFIFG));    //waits for the flag to set

  TBCCTL1 = CM_1 + CAP + CCIE;   //enable TimerB for capture
  TBCCTL2 = CM_1 + CAP + CCIE;   //enable TimerB Capture interrupt
  TBCCTL3 = CM_1 + CAP + CCIE;
  TBCCTL4 = CM_1 + CAP + CCIE;
  TBCCTL5 = CM_1 + CAP + CCIE;
  TBCCTL6 = CM_1 + CAP + CCIE;
  TBCTL = TBSSEL_0 + MC_2 + TBCLR + TBIE;
  EINT();
}

void magnetometer_communication_init(void){           //uart-initialization
  int ubr1_value;
  ubr1_value = F_CPU/(16*MAG_BAUDRATE)-1;
  
  WDTCTL = WDTPW + WDTHOLD;   //stopping WDT
  P3SEL |= 0x30;              //PORT3-4,5 = USART1 TX/RX
  ME2 |= UTXE1 + URXE1;       //Enable USART1 RX/TX
  UCTL1 |= CHAR;              //8bit character
  UTCTL1 |= SSEL0;
  U1BR0 = ubr1_value;         //51
  IE2 |= URXIE1;
}
