/*********************************************************
File:   efk.c
Author: Jan Rohde
Date:   4.6.07
Description:  Extended Kalman Filter functions
*********************************************************/

#include "ekf.h"

  /*
    Extended Kalman Filter equations used to obtain new sets of estimates
  */

void GetMeasurement(float **y_m, float **y_r){
  /*
  Sends a START signal to the magnetometer.
  Get lastest measurements from the sensors and sets the y_m pointer to point to them.
  When used with y_meas this matrix will point to the latest measurements obtained
  */
}
  
  /*
    GAUSS-NEWTON ALGORITHM, creating Jacobie-matrix and estimate new optimal quaternion, q_hat_next
  */
void GenerateJacobian(float **vPointers_ek[], Quaternion vQuaternions_ek[]){
  int i;
  float **dm_deta, **dm_deps1, **dm_deps2, **dm_deps3;
  float **T1, **T2, **T3, **T4;

  
  //creating dynamic matrices
  dm_deta = matrix(1, 6, 1, 6);
  dm_deps1 = matrix(1, 6, 1, 6);
  dm_deps2 = matrix(1, 6, 1, 6);
  dm_deps3 = matrix(1, 6, 1, 6);
  T1 = matrix(1, 6, 1, 1);
  T2 = matrix(1, 6, 1, 1);
  T3 = matrix(1, 6, 1, 1);
  T4 = matrix(1, 6, 1, 1);

  matrix_zeros(dm_deta, 6 ,6);
  matrix_zeros(dm_deps1, 6 ,6);
  matrix_zeros(dm_deps2, 6 ,6);
  matrix_zeros(dm_deps3, 6 ,6);
  matrix_zeros(T1, 6, 1);
  matrix_zeros(T2, 6, 1);
  matrix_zeros(T3, 6, 1);
  matrix_zeros(T4, 6, 1);
  
  vQuaternions_ek[0] = vQuaternions_ek[1];
  create_dm(dm_deta, dm_deps1, dm_deps2, dm_deps3, vQuaternions_ek[0]);
  matrix_multiplication(dm_deta, vPointers_ek[14], T1, 6, 6, 1);
  matrix_multiplication(dm_deps1, vPointers_ek[14], T2, 6, 6, 1);
  matrix_multiplication(dm_deps2, vPointers_ek[14], T3, 6, 6, 1);
  matrix_multiplication(dm_deps3, vPointers_ek[14], T4, 6, 6, 1);
  for(i = 1; i <= 6; i++){
    vPointers_ek[0][1][i] = -T1[1][i];
    vPointers_ek[0][2][i] = -T2[2][i];
    vPointers_ek[0][3][i] = -T2[3][i];
    vPointers_ek[0][4][i] = -T3[4][i];          //J = -[dm_deta*y_meas dm_deps1*y_meas dm_deps2*y_meas dm_deps3*y_meas]  (6 x 4)
  }
  //freeing up used temporary matrices from memory
  free_matrix(T1, 1, 6, 1, 1);
  free_matrix(T2, 1, 6, 1, 1);
  free_matrix(T3, 1, 6, 1, 1);
  free_matrix(T4, 1, 6, 1, 1);
}

  /*
    Creating M (6 x 6)
  */
void CreateM(float **vPointers_ek[], Quaternion vQuaternions_ek[]){
  int i,j;
  float **T1, **T2, **T3, **T4, **T5;
  float temp_value_1;

  T1 = matrix(1, 3, 1, 3);                                  //T1 = skew-symetric epsilon
  T2 = matrix(1, 3, 1, 3);
  T3 = matrix(1, 3, 1, 3);
  T4 = matrix(1, 3, 1, 3);
  T5 = matrix(1, 3, 1, 3);                                  //T5 == R_opt
  matrix_ident(T4, 3);
  skew_matrix_eps(vQuaternions_ek[0].eps, T1);
  matrix_multiplication(T1, T1, T2, 3, 3, 3);
  matrix_const_multiplication(T2, T3, 3, 3, 2.0f);          //2*skew^2(eps)
  temp_value_1 = 2.0f*vQuaternions_ek[0].eta;
  matrix_const_multiplication(T1, T2, 3, 3, temp_value_1);  //2*eta*skew(eps)
  matrix_addition(T2, T3, T1, 3, 3);                        //T1 = 2*eta*skew(eps) + 2*skew^2(eps)
  matrix_addition(T4, T1, T5, 3, 3);                        //T5 = eye(3) + 2*eta*skew(eps) + 2*skew^2(eps)
  for (i = 1; i <= 3; i++){
    for (j = 1; j <= 3; j++){
      vPointers_ek[1][i][j] = T5[i][j];
      vPointers_ek[1][i+3][j+3] = T5[i][j];
    }
  }
  //freeing up used temporary matrices from memory
  free_matrix(T1, 1, 3, 1, 3);
  free_matrix(T2, 1, 3, 1, 3);
  free_matrix(T3, 1, 3, 1, 3);
  free_matrix(T4, 1, 3, 1, 3);
  free_matrix(T5, 1, 3, 1, 3);
}

  /*
    creating next q_hat: q_hat_next = q_hat-inv(J'*J)*J'*(y_ref-M*y_meas)
  */
void Calculate_q_hat(float **vPointers_ek[], Quaternion vQuaternions_ek[]){
  float **T1, **T2, **T3, **T4, **T5;
  
  T1 = matrix(1, 4, 1, 6);
  T2 = matrix(1, 6, 1, 1);
  T3 = matrix(1, 6, 1, 1);
  T4 = matrix(1, 4, 1, 1);
  T5 = matrix(1, 4, 1, 4);
  matrix_transpose(vPointers_ek[0], T1, 6, 4);                             //T1 = J^T  (4 x 6)
  matrix_multiplication(vPointers_ek[1], vPointers_ek[14], T2, 6, 6, 1);
  matrix_subtraction(vPointers_ek[15], T2, T3, 6, 1);                      //y_ref - M*y_meas
  matrix_multiplication(T1, T3, T4, 4, 6, 1);                              //J^T*(y_ref - M*y_meas) (4 x 1)
  matrix_multiplication(T1, vPointers_ek[0], T5, 4, 6, 4);                 //(J^T*J) (4 x 4)
  free_matrix(T2, 1, 6, 1, 6);
  free_matrix(T3, 1, 6, 1, 6);
  T2 = matrix(1, 4, 1, 1);
  gaussj(T5, 4, NULL, 0);                                                  //(J^T*J)^-1
  matrix_multiplication(T5, T4, T2, 4, 4, 1);                              //(J^T*J)^-1*J^T*(y_ref - M*y_meas)
  quaternion_vec_subtraction(vQuaternions_ek[0], T2, vQuaternions_ek[1]);  //q_hat_next = q_hat - inv(J^T*J)*J^T*(y_ref - M*y_meas)
  free_matrix(T1, 1, 4, 1, 6);
  free_matrix(T2, 1, 4, 1, 4);
  free_matrix(T4, 1, 4, 1, 1);
  free_matrix(T5, 1, 4, 1, 4);
}
  /*
    EXTENDED KALMAN ALGORITHM - Computing Kalman Gain K
  */
void CreateK(float **vPointers_ek[], Quaternion vQuaternions_ek[]){
  int i,j;
  float **T1, **T2, **T3, **T4, **T5;

  vPointers_ek[12] = vPointers_ek[13];
  T1 = matrix(1, 6, 1, 3);
  T2 = matrix(1, 3, 1, 6);
  T3 = matrix(1, 3, 1, 3);
  T4 = matrix(1, 3, 1, 3);
  T5 = matrix(1, 6, 1, 3);
  matrix_transpose(vPointers_ek[4], T1, 3, 6);                              //T1 = H^T (6 x 3)
  matrix_multiplication(vPointers_ek[4], vPointers_ek[12], T2, 3, 6, 6);    //T2 = H*P_bar (3 x 6)
  matrix_multiplication(T2, T1, T3, 3, 6, 3);                               //T3 = H*P_bar*H^T (3 x 3)
  matrix_addition(T3, vPointers_ek[5], T4, 3, 3);                           //T4 = H*P_bar*H^T + R (3 x 3)
  gaussj(T4, 3, NULL, 0);                                                   //T4 = (H*P_bar*H^T + R)^-1 (3 x 3)
  matrix_multiplication(vPointers_ek[12], T1, T5, 6, 6, 3);                 //T5 = P_bar*H^T (6 x 3)
  matrix_multiplication(T5, T4, vPointers_ek[3], 6, 3, 3);                  //K = P_bar*H^T*(H*P_bar*H^T + R)^-1 (6 x 3)
  free_matrix(T1, 1, 6, 1, 3);
  free_matrix(T2, 1, 3, 1, 6);
  free_matrix(T3, 1 ,3, 1, 3);
  free_matrix(T4, 1 ,3, 1, 3);
  free_matrix(T5, 1, 6, 1, 3);
  /*
    divides K into Ke and Kw, Ke = upper half of K, Kw = lower half of K
  */
  for (i = 1; i <= 3; i++){
    for (j = 1; j <= 3; j++){
      vPointers_ek[8][i][j] = vPointers_ek[3][i][j];
      vPointers_ek[9][i][j] = vPointers_ek[3][i+3][j+3];
    }
  }
}

  /*
    Updating estimation error
  */
void UpdateEstimationError(float **vPointers_ek[], Quaternion vQuaternions_ek[]){
  vQuaternions_ek[3] = vQuaternions_ek[2];
  quat_prod(vQuaternions_ek[0], vQuaternions_ek[3], vQuaternions_ek[9], 1);                        //q_error = quaternionproduct(q_hat;q_bar)
  vQuaternions_ek[4].eps[0] = vQuaternions_ek[9].eps[0];
  vQuaternions_ek[4].eps[1] = vQuaternions_ek[9].eps[1];
  vQuaternions_ek[4].eps[2] = vQuaternions_ek[9].eps[2];
  vQuaternions_ek[6].eps[0] = vQuaternions_ek[0].eps[0] - vQuaternions_ek[3].eps[0];
  vQuaternions_ek[6].eps[1] = vQuaternions_ek[0].eps[1] - vQuaternions_ek[3].eps[1];
  vQuaternions_ek[6].eps[2] = vQuaternions_ek[0].eps[2] - vQuaternions_ek[3].eps[2];
}

  /*
    Updating state estimation
  */
void UpdateStateEstimation(float **vPointers_ek[], Quaternion vQuaternions_ek[]){
  int i;
  float temp;
  float **T1, **T2;

  if(vQuaternions_ek[9].eta < 0){
    i = -1;
  }
  else{
    i = 1;
  }

  //Updating quaternion estimation
  matrix_quat_multiplication(vPointers_ek[8], vQuaternions_ek[4], vQuaternions_ek[5]);  //Deps = Ke*deps
  temp = pow(vQuaternions_ek[5].eps[0], 2) + pow(vQuaternions_ek[5].eps[1], 2) + pow(vQuaternions_ek[5].eps[2], 2);
  vQuaternions_ek[7].eta = sqrt(1 - sqrt(temp)) * i;
  vQuaternions_ek[7].eps[0] = vQuaternions_ek[5].eps[0];
  vQuaternions_ek[7].eps[1] = vQuaternions_ek[5].eps[1];                        //placing Deps.eps in eta.eps for convenience on next operation
  vQuaternions_ek[7].eps[2] = vQuaternions_ek[5].eps[2];                        //eta = sqrt(1-norm(Deps,2)^2)*sign(q_error(1))
  quat_prod(vQuaternions_ek[7], vQuaternions_ek[3], vQuaternions_ek[8], 0);     //q_est updated

  //Updating omega estimation
  vPointers_ek[18] = vPointers_ek[16];
  T1 = matrix(1, 3, 1, 1);
  T2 = matrix(1, 3, 1, 1);
  T1[1][1] = vQuaternions_ek[6].eps[0];
  T1[2][1] = vQuaternions_ek[6].eps[1];
  T1[3][1] = vQuaternions_ek[6].eps[2];
  matrix_multiplication(vPointers_ek[9], T1, T2, 3, 3, 1);              //T2 = Kw*delta_eps_w
  matrix_addition(vPointers_ek[18], T2, vPointers_ek[19], 3, 1);        //w_est = w_bar + Kw*delta_eps_w
  free_matrix(T1, 1, 3, 1, 1);
  free_matrix(T2, 1, 3, 1, 1);
  vPointers_ek[20][1][1] = vQuaternions_ek[8].eta;
  vPointers_ek[20][2][1] = vQuaternions_ek[8].eps[0];
  vPointers_ek[20][3][1] = vQuaternions_ek[8].eps[1];
  vPointers_ek[20][4][1] = vQuaternions_ek[8].eps[2];
  vPointers_ek[20][5][1] = vPointers_ek[19][1][1];
  vPointers_ek[20][6][1] = vPointers_ek[19][2][1];
  vPointers_ek[20][7][1] = vPointers_ek[19][3][1];                       //x_est = q_est; w_est (7 x 1)
}

  /*
    Updating Error Covariance P (6 x 6)
  */
void UpdateErrorCovariance(float **vPointers_ek[], Quaternion vQuaternions_ek[]){
  float **T1, **T2, **T3, **T4, **T5, **T6;

  T1 = matrix(1, 6, 1, 6);
  T2 = matrix(1, 6, 1, 6);
  T3 = matrix(1, 6, 1, 6);
  T4 = matrix(1, 6, 1, 6);
  T5 = matrix(1, 6, 1, 6);
  T6 = matrix(1, 6, 1, 6);
  matrix_zeros(T1, 6, 6);                                                 //T1 = eye(6)
  matrix_multiplication(vPointers_ek[3], vPointers_ek[4], T2, 6, 3, 6);   //T2 = K*H
  matrix_subtraction(T1, T2, T3, 6, 6);                                   //T3 = eye(6) - K*H
  matrix_transpose(T3, T4, 6, 6);                                         //T4 = (eye(6) - K*H)^T
  matrix_multiplication(T3, vPointers_ek[12], T5, 6, 6, 6);               //T5 = (eye(6) - K*H)*P_bar
  matrix_multiplication(T5, T4, T6, 6, 6, 6);                             //T6 = (eye(6) - K*H)*P_bar*(eye(6) - K*H)^T
  free_matrix(T1, 1, 6, 1, 6);
  free_matrix(T2, 1, 6, 1, 6);
  free_matrix(T3, 1, 6, 1, 6);
  free_matrix(T4, 1, 6, 1, 6);
  T1 = matrix(1, 3, 1, 6);
  T2 = matrix(1, 6, 1, 3);
  matrix_transpose(vPointers_ek[3], T1, 6, 3);                            //T1 = K^T (3 x 6)
  matrix_multiplication(vPointers_ek[3], vPointers_ek[5], T2, 6, 3, 3);   //T2 = K*R (6 x 3)
  matrix_multiplication(T2, T1, T5, 6, 3, 6);                             //T5 = K*R*K^T (6 x 6)
  matrix_addition(T6, T5, vPointers_ek[2], 6, 6);                         //P = (eye(6) - K*H)*P_bar*(eye(6) - K*H)^T + K*R*K^T
  free_matrix(T1, 1, 3, 1, 6);
  free_matrix(T2, 1, 6, 1, 3);
  free_matrix(T5, 1, 6, 1, 6);
  free_matrix(T6, 1, 6, 1, 6);
}
  /*
    Forward Euler Integration: Finding Phi_r and Phi_tot
    Phi_r = eye(6) + linearsystemR(x_est);
    Phi_tot = nonLinearProp(x_est);
    x_est = [eta_hat eps_1_hat eps_2_hat eps_3_hat w_est_x w_est_y w_est_z]
  */
void UpdateStateTransitionMatrices(float **vPointers_ek[], Quaternion vQuaternions_ek[]){
  int i,j;
  float w_o = -1.083e-3;                                      //sqrt((G*M)/R^3)
  float temp_value;
  float **T1, **T2, **T3, **T4, **T5, **T6, **T7, **T8;

  //Phi_r = eye(6) + linearsystemR(x_est);
  T1 = matrix(1, 3, 1, 3);
  T2 = matrix(1, 3, 1, 3);
  T3 = matrix(1, 3, 1, 3);
  T4 = matrix(1, 3, 1, 3);
  matrix_ident(T3, 3);
  skew_matrix(vPointers_ek[19], T1);
  matrix_const_multiplication(T1, T1, 3, 3, -0.5f);                         //T1 = -1/2*skew(w_est)
  matrix_const_multiplication(T3, T3, 3, 3, vPointers_ek[20][1][1]);        //T3 = eta*eye(3)
  skew_matrix_eps(vQuaternions_ek[8].eps, T4);                              //T4 = skew(eps)
  matrix_addition(T3, T4, T3, 3, 3);                                        //T3 = eta*eye(3) + skew(eps)
  matrix_const_multiplication(T3, T3, 3, 3, 0.5f);                          //T3 = 1/2*(eta*eye(3) + skew(eps))
  for(i = 1; i <= 3; i++){
    for(j = 1; j <= 3; j++){
      vPointers_ek[24][i][j] = T1[i][j];
      vPointers_ek[24][i][j+3] = T3[i][j];
    }
  }                                                                         //F_ATT = [-1/2*skew(w_est) 1/2*(eta*eye(3)+skew(eps))]
  matrix_ident(T3, 3);                                                      //T3 = eye(3)
  skew_matrix_eps(vQuaternions_ek[8].eps, T4);                              //T4 = skew(eps)
  matrix_multiplication(T4, T4, T2, 3, 3, 3);                               //T2 = skew(eps)^2
  temp_value = 2*vQuaternions_ek[8].eta;
  matrix_const_multiplication(T4, T4, 3, 3, temp_value);                  //T4 = 2*eta*skew(eps)
  matrix_const_multiplication(T2, T2, 3, 3, 2.0f);                          //T2 = skew(eps)^2
  matrix_addition(T3, T4, T3, 3, 3);
  matrix_addition(T3, T2, T1, 3, 3);                                        //T1 = R_B_O
  free_matrix(T2, 1, 3, 1, 3);
  free_matrix(T3, 1, 3, 1, 3);
  free_matrix(T4, 1, 3, 1, 3);
  T2 = matrix(1, 3, 1, 1);
  T2[1][1] = T1[1][2];
  T2[2][1] = T1[2][2];
  T2[3][1] = T1[3][2];                                                      //T2 =
  matrix_const_multiplication(T2, T2, 3, 1, w_o);
  matrix_addition(vPointers_ek[19], T2, T2, 3, 1);                          //T2 = w = w_est + R_B_O[;,2]*w_o (3 x 1)
  create_linearsysR_elements(T1, T2, vPointers_ek[20], vPointers_ek[25], vPointers_ek[26], w_o);   //F_VEL = [a52 a53 ...; a62 a63 ...; a72 a73 ...]
  T3 = matrix(1, 6, 1, 6);
  T4 = matrix(1, 6, 1, 6);
  matrix_ident(T4, 6);
  /*
    Creating F = [F_ATT; F_VEL]  for Phi_r
  */
  for(i = 1; i <= 3; i++){
    for(j = 1; j <= 6; j++){
      T3[i][j] = vPointers_ek[24][i][j];
      T3[i+3][j] = vPointers_ek[25][i][j];
    }
  }

  matrix_addition(T4, T3, vPointers_ek[10], 6, 6);                          //Phi_r = eye(6) + F
  free_matrix(T2, 1, 3, 1, 1);
  free_matrix(T3, 1, 6, 1, 6);
  free_matrix(T4, 1, 6, 1, 6);

  /*
    Creating F = [F_ATT; F_VEL] for Phi_tot
  */
  T2 = matrix(1, 3, 1, 1);
  T3 = matrix(1, 1, 1, 3);
  T4 = matrix(1, 1, 1, 1);
  T2[1][1] = vPointers_ek[20][2][1];
  T2[2][1] = vPointers_ek[20][3][1];
  T2[3][1] = vPointers_ek[20][4][1];                        //T2 = x_est_epsilon
  matrix_transpose(T2, T3, 3, 1);                           //T3 = x_est_epsilon^T
  matrix_multiplication(vPointers_ek[19], T3, T4, 1, 3, 1); //T4 = w_est*x_est_epsilon^T
  matrix_const_multiplication(T4, T4, 1, 1, -0.5f);
  vPointers_ek[11][1][1] = T4[1][1];
  free_matrix(T2, 1, 3, 1, 1);
  free_matrix(T3, 1, 1, 1, 3);
  free_matrix(T4, 1, 1, 1, 1);
  T2 = matrix(1, 3, 1, 3);
  T3 = matrix(1, 3, 1, 3);
  T4 = matrix(1, 3, 1, 1);
  matrix_ident(T2, 3);
  matrix_const_multiplication(T2, T2, 3, 3, vPointers_ek[20][1][1]);   //T4 = eta*eye(3)
  skew_matrix_eps(vQuaternions_ek[8].eps, T3);                              //T3 = skew(eps)
  matrix_addition(T2, T3, T3, 3, 3);                        //T3 = eta*eye(3) + skew(eps)
  matrix_const_multiplication(T3, T3, 3, 3, 0.5f);          //T3 = 0.5*(eta*eye(3) + skew(eps))
  matrix_multiplication(T3, vPointers_ek[19], T4, 3, 3, 1);            //T4 = 0.5*(eta*eye(3) + skew(eps))*w_est
  vPointers_ek[11][2][1] = T4[1][1];                                   //F_attitude is now estimated
  vPointers_ek[11][3][1] = T4[2][1];
  vPointers_ek[11][4][1] = T4[3][1];
  free_matrix(T2, 1, 3, 1, 3);
  free_matrix(T3, 1, 3, 1, 3);
  free_matrix(T4, 1, 3, 1, 1);

  T2 = matrix(1, 3, 1, 1);
  T3 = matrix(1, 3, 1, 3);
  T4 = matrix(1, 3, 1, 3);
  T5 = matrix(1, 3, 1, 1);
  T6 = matrix(1, 3, 1, 3);
  T7 = matrix(1, 3, 1, 1);
  T8 = matrix(1, 3, 1, 1);

  T2[1][1] = 0.0f;
  T2[2][1] = w_o;
  T2[3][1] = 0.0f;                                          //T2 = [0.0 w_o 0.0]^T
  skew_matrix(vPointers_ek[19], T3);
  matrix_multiplication(T3, T1, T4, 3, 3, 3);
  matrix_multiplication(T4, T2, T5, 3, 3, 1);              //T5 = skew(w_est)*RBO*[0.0 w_o 0.0]^T (3 x 1)
  free_matrix(T3, 1, 3, 1, 3);
  free_matrix(T4, 1, 3, 1, 3);
  T3 = matrix(1, 3, 1, 1);
  T4 = matrix(1, 3, 1, 1);
  matrix_multiplication(T1, T2, T3, 3, 3, 1);
  matrix_addition(vPointers_ek[19], T3, T3, 3, 1);                     //T3 = w_est+RBO*[0.0 w_o 0.0]^T    (3 x 1)
  matrix_multiplication(vPointers_ek[27], T3, T4, 3, 1, 1);                //T4 = I*(w_est+RBO*[0.0 w_o 0.0]^T) (3 x 1)
  skew_matrix(T3, T6);                                     //T6 = skew(w_est+RBO*[0.0 w_o 0.0]^T)
  matrix_const_multiplication(T6, T6, 3, 3, -1.0f);       //T6 = -skew(w_est+RBO*[0.0 w_o 0.0]^T)
  matrix_multiplication(T6, T4, T7, 3, 3, 1);             //T7 = -skew(w_est+RBO*[0.0 w_o 0.0]^T)*I*(w_est+RBO*[0.0 w_o 0.0]^T) (3 x 1)
  copy_matrix(vPointers_ek[27], T6, 1, 3, 1, 3);                          //T6 = I
  gaussj(T6, 3, NULL, 0);                                  //returns inv(T6) = inv(I)
  matrix_multiplication(T6, T7, T8, 3, 3, 1);            //T8 = inv(I)*(-skew(w_est+RBO*[0.0 w_o 0.0]^T)*I*(w_est+RBO*[0.0 w_o 0.0]^T))
  matrix_addition(T8, T5, T8, 3, 1);                     //T8 = inv(I)*(-skew(w_est+RBO*[0.0 w_o 0.0]^T)*I*(w_est+RBO*[0.0 w_o 0.0]^T)) + T10 = skew(w_est)*RBO*[0.0 w_o 0.0]^T
  vPointers_ek[11][5][1] = T8[1][1];
  vPointers_ek[11][6][1] = T8[2][1];
  vPointers_ek[11][7][1] = T8[3][1];                      //Phi_t created
  free_matrix(T1, 1, 3, 1, 3);
  free_matrix(T2, 1, 3, 1, 1);
  free_matrix(T3, 1, 3, 1, 1);
  free_matrix(T4, 1, 3, 1, 1);
  free_matrix(T5, 1, 3, 1, 1);
  free_matrix(T6, 1, 3, 1, 3);
  free_matrix(T7, 1, 3, 1, 1);
  free_matrix(T8, 1, 3, 1, 1);
}

  /*
    state estimation propagation
  */
void StatePropagation(float **vPointers_ek[], Quaternion vQuaternions_ek[]){
  float **T1, **T2, **T3, **T4;

  matrix_addition(vPointers_ek[11], vPointers_ek[20], vPointers_ek[17], 7, 1);//x_bar_next = Phi_tot + x_est
  vQuaternions_ek[2].eta = vPointers_ek[17][1][1];
  vQuaternions_ek[2].eps[0] = vPointers_ek[17][2][1];
  vQuaternions_ek[2].eps[1] = vPointers_ek[17][3][1];
  vQuaternions_ek[2].eps[2] = vPointers_ek[17][4][1];                         //q_bar_next = x_bar_next(1:4)
  vPointers_ek[16][1][1] = vPointers_ek[17][5][1];
  vPointers_ek[16][2][1] = vPointers_ek[17][6][1];
  vPointers_ek[16][3][1] = vPointers_ek[17][7][1];                            //w_bar_next = x_bar_next(5:7)
  vQuaternions_ek[2].eta = vQuaternions_ek[2].eta/sqrt(pow(vQuaternions_ek[2].eta, 2) + pow(vQuaternions_ek[2].eps[0], 2) + pow(vQuaternions_ek[2].eps[1], 2) + pow(vQuaternions_ek[2].eps[2], 2));
  vQuaternions_ek[2].eps[0] = vQuaternions_ek[2].eps[0]/sqrt(pow(vQuaternions_ek[2].eta, 2) + pow(vQuaternions_ek[2].eps[0], 2) + pow(vQuaternions_ek[2].eps[1], 2) + pow(vQuaternions_ek[2].eps[2], 2));
  vQuaternions_ek[2].eps[1] = vQuaternions_ek[2].eps[1]/sqrt(pow(vQuaternions_ek[2].eta, 2) + pow(vQuaternions_ek[2].eps[0], 2) + pow(vQuaternions_ek[2].eps[1], 2) + pow(vQuaternions_ek[2].eps[2], 2));
  vQuaternions_ek[2].eps[2] = vQuaternions_ek[2].eps[2]/sqrt(pow(vQuaternions_ek[2].eta, 2) + pow(vQuaternions_ek[2].eps[0], 2) + pow(vQuaternions_ek[2].eps[1], 2) + pow(vQuaternions_ek[2].eps[2], 2));
  vPointers_ek[17][1][1] = vQuaternions_ek[2].eta;
  vPointers_ek[17][2][1] = vQuaternions_ek[2].eps[0];
  vPointers_ek[17][3][1] = vQuaternions_ek[2].eps[1];
  vPointers_ek[17][4][1] = vQuaternions_ek[2].eps[2];

  //Error covariance propagation
  T1 = matrix(1, 6, 1, 6);
  T2 = matrix(1, 6, 1, 6);
  T3 = matrix(1, 6, 1, 6);
  T4 = matrix(1, 6, 1, 6);
  matrix_transpose(vPointers_ek[6], T1, 6, 6);                              //T1 = G^T
  matrix_multiplication(vPointers_ek[6], vPointers_ek[7], T2, 6, 6, 6);     //T2 = G*Q
  matrix_multiplication(T2, T1, T3, 6, 6, 6);                               //T3 = G*Q*G^T
  matrix_transpose(vPointers_ek[10], T1, 6, 6);                             //T1 = (Phi_r)^T
  matrix_multiplication(vPointers_ek[10], vPointers_ek[2], T2, 6, 6, 6);    //T2 = Phi_r*P
  matrix_multiplication(T2, T1, T4, 6, 6, 6);                               //T4 = Phi_r*P*(Phi_r)^T
  matrix_addition(T4, T3, vPointers_ek[12], 6, 6);                          //P_bar_next = Phi_r*P*(Phi_r)^T + G*Q*G^T
  vPointers_ek[21][1][1] = vPointers_ek[20][1][1];
  vPointers_ek[21][2][1] = vPointers_ek[20][2][1];
  vPointers_ek[21][3][1] = vPointers_ek[20][3][1];
  vPointers_ek[21][4][1] = vPointers_ek[20][4][1];
  vPointers_ek[21][5][1] = vPointers_ek[20][5][1];
  vPointers_ek[21][6][1] = vPointers_ek[20][6][1];
  vPointers_ek[21][7][1] = vPointers_ek[20][7][1];
  free_matrix(T1, 1, 6, 1, 6);
  free_matrix(T2, 1, 6, 1, 6);
  free_matrix(T3, 1, 6, 1, 6);
  free_matrix(T4, 1, 6, 1, 6);
}

  /*
    Transmit estimates
  */
void TransmitEstimates(float **X_hat){

}

  /*
    Creating F_VEL = [a52 a53 a54 0 a56 a57
                      a62 a63 a64 a65 0 a67
                      a72 a73 a74 a75 a76 0]
  */
void create_linearsysR_elements(float **RBO, float **w, float **x_est, float **A, float **K, float w_o){
  float a52, a53, a54, a56, a57;
  float a62, a63, a64, a65, a67;
  float a72, a73, a74, a75, a76;

  a52 = 2*K[1][1]*w_o*(x_est[1][1]*w[2][1] + x_est[2][1]*w[3][1]) + 6*K[1][1]*pow(w_o,2)*(x_est[2][1]*RBO[2][3] - x_est[1][1]*RBO[3][3]) + 2*(x_est[1][1]*x_est[6][1] - x_est[2][1]*x_est[7][1])*w_o;
  a53 = -2*K[1][1]*w_o*(x_est[4][1]*w[2][1] + x_est[3][1]*w[3][1]) + 6*K[1][1]*pow(w_o,2)*(x_est[3][1]*RBO[2][3] - x_est[4][1]*RBO[3][3]) + 2*(x_est[3][1]*x_est[7][1] - x_est[4][1]*x_est[6][1])*w_o;
  a54 = 2*K[1][1]*w_o*(x_est[4][1]*w[3][1] + x_est[3][1]*w[2][1]) - 6*K[1][1]*pow(w_o,2)*(x_est[3][1]*RBO[3][3] - x_est[4][1]*RBO[2][3]) - 2*(x_est[4][1]*x_est[7][1] - x_est[3][1]*x_est[6][1])*w_o;
  a56 = K[1][1]*w[3][1] - RBO[3][2]*w_o;
  a57 = K[1][1]*w[2][1] - RBO[2][2]*w_o;

  a62 = 2*K[2][1]*w_o*(x_est[3][1]*w[3][1] - x_est[1][1]*w[1][1]) + 6*K[2][1]*pow(w_o,2)*(x_est[4][1]*RBO[3][3] - x_est[2][1]*RBO[1][3]) - 2*(x_est[4][1]*x_est[5][1] + x_est[3][1]*x_est[7][1])*w_o;
  a63 = 2*K[2][1]*w_o*(x_est[2][1]*w[3][1] + x_est[4][1]*w[1][1]) - 6*K[2][1]*pow(w_o,2)*(x_est[1][1]*RBO[3][3] + x_est[3][1]*RBO[1][3]) + 2*(x_est[4][1]*x_est[5][1] - x_est[2][1]*x_est[7][1])*w_o;
  a64 = 2*K[2][1]*w_o*(x_est[1][1]*w[3][1] - x_est[3][1]*w[1][1]) + 6*K[2][1]*pow(w_o,2)*(x_est[2][1]*RBO[3][3] + x_est[4][1]*RBO[1][3]) + 2*(x_est[3][1]*x_est[5][1] - x_est[1][1]*x_est[7][1])*w_o;
  a65 = -K[2][1]*w[3][1] + RBO[3][2]*w_o;
  a67 = -K[2][1]*w[1][1] - RBO[1][2]*w_o;

  a72 = 2*K[3][1]*w_o*(x_est[3][1]*w[2][1] - x_est[2][1]*w[1][1]) + 6*K[3][1]*pow(w_o,2)*(x_est[4][1]*RBO[2][3] + x_est[1][1]*RBO[1][3]) + 2*(x_est[2][1]*x_est[5][1] + x_est[3][1]*x_est[6][1])*w_o;
  a73 = 2*K[3][1]*w_o*(x_est[2][1]*w[2][1] + x_est[3][1]*w[1][1]) + 6*K[3][1]*pow(w_o,2)*(x_est[4][1]*RBO[1][3] + x_est[1][1]*RBO[2][3]) + 2*(x_est[2][1]*x_est[6][1] - x_est[3][1]*x_est[5][1])*w_o;
  a74 = 2*K[3][1]*w_o*(x_est[1][1]*w[2][1] - x_est[4][1]*w[1][1]) + 6*K[3][1]*pow(w_o,2)*(x_est[2][1]*RBO[2][3] + x_est[3][1]*RBO[1][3]) + 2*(x_est[4][1]*x_est[5][1] + x_est[1][1]*x_est[6][1])*w_o;
  a75 = -K[3][1]*w[2][1] - RBO[2][2]*w_o;
  a76 = -K[3][1]*w[1][1] + RBO[1][2]*w_o;

  A[1][1] = a52;
  A[1][2] = a53;
  A[1][3] = a54;
  A[1][4] = 0.0f;
  A[1][5] = a56;
  A[1][6] = a57;
  
  A[2][1] = a62;
  A[2][2] = a63;
  A[2][3] = a64;
  A[2][4] = a65;
  A[2][5] = 0.0f;
  A[2][6] = a67;

  A[3][1] = a72;
  A[3][2] = a73;
  A[3][3] = a74;
  A[3][4] = a75;
  A[3][5] = a76;
  A[3][6] = 0.0f;
}

  /*
    Creating partial derivative matrices for J
  */
static void create_dm(float **dmdeta, float **dmdeps1, float **dmdeps2, float **dmdeps3, Quaternion quater){
  dmdeta[1][1] = 2*quater.eta;
  dmdeta[1][2] = 2*-quater.eps[2];
  dmdeta[1][3] = 2*quater.eps[1];
  dmdeta[2][1] = 2*quater.eps[2];
  dmdeta[2][2] = 2*quater.eta;
  dmdeta[2][3] = 2*-quater.eps[0];
  dmdeta[3][1] = 2*-quater.eps[1];
  dmdeta[3][2] = 2*quater.eps[0];
  dmdeta[3][3] = 2*quater.eta;
  dmdeta[4][4] = 2*quater.eta;
  dmdeta[4][5] = 2*-quater.eps[2];
  dmdeta[4][6] = 2*quater.eps[1];
  dmdeta[5][4] = 2*quater.eps[2];
  dmdeta[5][5] = 2*quater.eta;
  dmdeta[5][6] = 2*-quater.eps[0];
  dmdeta[6][4] = 2*-quater.eps[1];
  dmdeta[6][5] = 2*quater.eps[0];
  dmdeta[6][6] = 2*quater.eta;
  
  dmdeps1[1][1] = 2*quater.eps[0];
  dmdeps1[1][2] = 2*quater.eps[1];
  dmdeps1[1][3] = 2*quater.eps[2];
  dmdeps1[2][1] = 2*quater.eps[1];
  dmdeps1[2][2] = 2*-quater.eps[0];
  dmdeps1[2][3] = 2*quater.eta;
  dmdeps1[3][1] = 2*quater.eps[2];
  dmdeps1[3][2] = 2*quater.eta;
  dmdeps1[3][3] = 2*-quater.eps[0];
  dmdeps1[4][4] = 2*quater.eps[0];
  dmdeps1[4][5] = 2*quater.eps[1];
  dmdeps1[4][6] = 2*quater.eps[2];
  dmdeps1[5][4] = 2*quater.eps[1];
  dmdeps1[5][5] = 2*-quater.eps[0];
  dmdeps1[5][6] = 2*quater.eta;
  dmdeps1[6][4] = 2*quater.eps[2];
  dmdeps1[6][5] = 2*quater.eta;
  dmdeps1[6][6] = 2*-quater.eps[0];
  
  dmdeps2[1][1] = 2*-quater.eps[1];
  dmdeps2[1][2] = 2*quater.eps[0];
  dmdeps2[1][3] = 2*quater.eta;
  dmdeps2[2][1] = 2*quater.eps[0];
  dmdeps2[2][2] = 2*quater.eps[1];
  dmdeps2[2][3] = 2*quater.eps[2];
  dmdeps2[3][1] = 2*-quater.eta;
  dmdeps2[3][2] = 2*quater.eps[2];
  dmdeps2[3][3] = 2*-quater.eps[1];
  dmdeps2[4][4] = 2*-quater.eps[1];
  dmdeps2[4][5] = 2*quater.eps[0];
  dmdeps2[4][6] = 2*quater.eta;
  dmdeps2[5][4] = 2*quater.eps[0];
  dmdeps2[5][5] = 2*quater.eps[1];
  dmdeps2[5][6] = 2*quater.eps[2];
  dmdeps2[6][4] = 2*-quater.eta;
  dmdeps2[6][5] = 2*quater.eps[2];
  dmdeps2[6][6] = 2*-quater.eps[1];

  dmdeps3[1][1] = 2*-quater.eps[2];
  dmdeps3[1][2] = 2*-quater.eta;
  dmdeps3[1][3] = 2*quater.eps[0];
  dmdeps3[2][1] = 2*quater.eta;
  dmdeps3[2][2] = 2*-quater.eps[2];
  dmdeps3[2][3] = 2*quater.eps[1];
  dmdeps3[3][1] = 2*quater.eps[0];
  dmdeps3[3][2] = 2*quater.eps[1];
  dmdeps3[3][3] = 2*quater.eps[2];
  dmdeps3[4][4] = 2*-quater.eps[2];
  dmdeps3[4][5] = 2*-quater.eta;
  dmdeps3[4][6] = 2*quater.eps[0];
  dmdeps3[5][4] = 2*quater.eta;
  dmdeps3[5][5] = 2*-quater.eps[2];
  dmdeps3[5][6] = 2*quater.eps[1];
  dmdeps3[6][4] = 2*quater.eps[0];
  dmdeps3[6][5] = 2*quater.eps[1];
  dmdeps3[6][6] = 2*quater.eps[2];
}
