/*********************************************************
File:   matirx_operations.c
Author: Jan Rohde
Date:   4.6.07
Description:  Matrix operations used in the
              Extended Kalman Filter.
*********************************************************/

#include "matrix_operations.h"

#define NR_END 1
#define FREE_ARG char*
#define SWAP(a,b) {temp=(a);(a)=(b);(b)=temp;}

/*
  copy a matrix/submatrix: A(n x m) -> B(n x m), rows n to N and columns m to M
*/
static void copy_matrix(float **A, float **B, int n, int N, int m, int M){
  int i, j;
  for (i = n; i <= N; i++){
    for (j = m; j <= M; j++){
      B[i][j] = A[i][j];
    }
  }
}

/*
  add two matrices with the same dimensions: A(n x m), B(n x m) -> C(n x m)
*/
static void matrix_addition(float **A, float **B, float **C, int N, int M){
  int i, j;
  for (i = 1; i <= N; i++){
    for (j = 1; j <= M; j++){
      C[i][j] = A[i][j] + B[i][j];
    }
  }
}

/*
  subtract two matrices with the same dimensions: A(n x m), B(n x m) -> C(n x m)
*/
static void matrix_subtraction(float **A, float **B, float **C, int N, int M){
  int i, j;
  for (i = 1; i <= N; i++){
    for (j = 1; j <= M; j++){
      C[i][j] = A[i][j] - B[i][j];
    }
  }
}

/*
  multiply two matrices: A(n x p), B(p x m) -> C(n x m)
*/
static void matrix_multiplication(float **A, float **B, float **C, int N, int P, int M){
  int i, j, k;
  for (i = 1; i <= N; i++){
    for (j = 1; j <= M; j++){
      C[i][j] = 0.0f;                           //force compiler to read number as float
      for (k = 1; k <= P; k++){
        C[i][j] += A[i][k]*B[k][j];
      }
    }
  }
}

/*
  transpose a matrix: A(n x m) -> B(m x n)
*/
static void matrix_transpose(float **A, float **B, int N, int M){
  int i, j;
  for (i = 1; i <= N; i++){
    for (j = 1; j <= M; j++){
      B[j][i] = A[i][j];
    }
  }
}

/*
  create a (n x m) matrix with all zeros
*/
static void matrix_zeros(float **A, int N, int M){
  int i, j;
  for (i = 1; i <= N; i++){
    for (j = 1; j <= M; j++){
      A[i][j] = 0.0f;           //force compiler to read number as float
    }
  }
}

/*
  create a (n x n) identity matrix
*/
static void matrix_ident(float **A, int N){
  int i, j;
  for (i = 1; i <= N; i++){
    for (j = 1; j <= N; j++){
      if(i == j)
        A[i][j] = 1.0f;           //force compiler to read number as float
      else
        A[i][j] = 0.0f;
    }
  }
}

/*
  create a (n x n) diag matrix
*/
static void matrix_diag(float **A, float *dv, int N){
  int i, j;
  for (i = 1; i <= N; i++){
    for (j = 1; j <= N; j++){
      if(i == j)
        A[i][j] = 1.0f * dv[i];           //force compiler to read number as float
      else
        A[i][j] = 0.0f;
    }
  }
}

/*
  Multiply a matrix with a constant a: A(n x m) * a -> B(n x m)
*/
static void matrix_const_multiplication(float **A, float **B, int N, int M, float a){
  int i, j;
  for (i = 1; i <= N; i++){
    for (j = 1; j <= M; j++){
      B[i][j] = A[i][j] * a;
    }
  }
}

/*
  create a skew matrix from epsilon
*/
static void skew_matrix_eps(float *vec, float **skew){
  skew[1][1] = 0.0f;
  skew[1][2] = -vec[2];
  skew[1][3] = vec[1];
  skew[2][1] = vec[2];
  skew[2][2] = 0.0f;
  skew[2][3] = -vec[0];
  skew[3][1] = -vec[1];
  skew[3][2] = vec[0];
  skew[3][3] = 0.0f;
}

/*
  create a skew matrix
*/
static void skew_matrix(float **vec, float **skew){
  skew[1][1] = 0.0f;
  skew[1][2] = -vec[3][1];
  skew[1][3] = vec[2][1];
  skew[2][1] = vec[3][1];
  skew[2][2] = 0.0f;
  skew[2][3] = -vec[1][1];
  skew[3][1] = -vec[2][1];
  skew[3][2] = vec[1][1];
  skew[3][3] = 0.0f;
}

/*
  compute the quaternion product: q1 x q2 -> q
*/
static void quat_prod(Quaternion q1, Quaternion q2, Quaternion q, int k){
  int i,j;
  float temp = 0;
  float **temp_eps_skew;

  if(k == 1){                               //if == 1, perform inverst quaternion product
    q2.eps[0] = -1*q2.eps[0];
    q2.eps[1] = -1*q2.eps[1];
    q2.eps[2] = -1*q2.eps[2];
  }

  temp_eps_skew = matrix(1, 3, 1, 3);
  skew_matrix_eps(q1.eps, temp_eps_skew);   //create temp_eps_skew matrix

  for(i = 1; i <= 3; i++){
    temp += q1.eps[i]*q2.eps[i];
  }
  q.eta = q1.eta * q2.eta - temp;   //new eta calculated

  for (i = 0; i < 3; i++){
    temp = 0;
    for (j = 0; j < 3; j++){
      temp += temp_eps_skew[i][j] * q2.eps[j];
    }
    q.eps[i] = q1.eta * q2.eps[i] + q2.eta * q1.eps[i] + temp;
  }
  free_matrix(temp_eps_skew, 1, 3, 1, 3);
  
  if(k == 1){                               //restore q2
    q2.eps[0] = -1*q2.eps[0];
    q2.eps[1] = -1*q2.eps[1];
    q2.eps[2] = -1*q2.eps[2];
  }
}

/*
  Multiply between a float matrix and a quaternion of type Quaternion
*/
static void matrix_quat_multiplication(float **A, Quaternion q1, Quaternion q2){
  int i, j;
  for(i = 1; i <= 3; i++){
    q2.eps[i-1] = 0.0f;
    for(j = 1; j <= 3; j++){
      q2.eps[i-1] += A[i][j]*q1.eps[j-1];
    }
  }
}

static void quaternion_vec_subtraction(Quaternion q1, float **A, Quaternion q2){
  int i;
  q2.eta = q1.eta - A[1][1];
  for (i = 1; i <= 3; i++){
    q2.eps[i-1] = q1.eps[i-1] - A[1][i+1];
  }
}

/*
  From numerical recipes
*/

float *vector(long nl, long nh)
/* allocate a float vector with subscript range v[nl..nh] */
{
	float *v;

	v=(float *)malloc((size_t) ((nh-nl+1+NR_END)*sizeof(float)));
	
	return v-nl+NR_END;
}

int *ivector(long nl, long nh)
/* allocate an int vector with subscript range v[nl..nh] */
{
	int *v;

	v=(int *)malloc((size_t) ((nh-nl+1+NR_END)*sizeof(int)));
	
	return v-nl+NR_END;
}

float **matrix(long nrl, long nrh, long ncl, long nch)
/* allocate a float matrix with subscript range m[nrl..nrh][ncl..nch] */
{
	long i, nrow=nrh-nrl+1,ncol=nch-ncl+1;
	float **m;

	/* allocate pointers to rows */
	m=(float **) malloc((size_t)((nrow+NR_END)*sizeof(float*)));
	
	m += NR_END;
	m -= nrl;

	/* allocate rows and set pointers to them */
	m[nrl]=(float *) malloc((size_t)((nrow*ncol+NR_END)*sizeof(float)));
	
	m[nrl] += NR_END;
	m[nrl] -= ncl;

	for(i=nrl+1;i<=nrh;i++) m[i]=m[i-1]+ncol;

	/* return pointer to array of pointers to rows */
	return m;
}

void free_vector(float *v, long nl, long nh)
/* free a float vector allocated with vector() */
{
	free((FREE_ARG) (v+nl-NR_END));
}

void free_ivector(int *v, long nl, long nh)
/* free an int vector allocated with ivector() */
{
	free((FREE_ARG) (v+nl-NR_END));
}

void free_matrix(float **m, long nrl, long nrh, long ncl, long nch)
/* free a float matrix allocated by matrix() */
{
	free((FREE_ARG) (m[nrl]+ncl-NR_END));
	free((FREE_ARG) (m+nrl-NR_END));
}

void gaussj(float **a, int n, float **b, int m){
	int *indxc,*indxr,*ipiv;
	int i,icol,irow,j,k,l,ll;
	float big,dum,pivinv,temp;

	indxc=ivector(1,n);
	indxr=ivector(1,n);
	ipiv=ivector(1,n);
	for (j=1;j<=n;j++) ipiv[j]=0;
	for (i=1;i<=n;i++) {
		big=0.0;
		for (j=1;j<=n;j++)
			if (ipiv[j] != 1)
				for (k=1;k<=n;k++) {
					if (ipiv[k] == 0) {
						if (fabs(a[j][k]) >= big) {
							big=fabs(a[j][k]);
							irow=j;
							icol=k;
						}
					}
				}
		++(ipiv[icol]);
		if (irow != icol) {
			for (l=1;l<=n;l++) SWAP(a[irow][l],a[icol][l])
			for (l=1;l<=m;l++) SWAP(b[irow][l],b[icol][l])
		}
		indxr[i]=irow;
		indxc[i]=icol;
//		if (a[icol][icol] == 0.0) nrerror("gaussj: Singular Matrix");
		pivinv=1.0/a[icol][icol];
		a[icol][icol]=1.0;
		for (l=1;l<=n;l++) a[icol][l] *= pivinv;
		for (l=1;l<=m;l++) b[icol][l] *= pivinv;
		for (ll=1;ll<=n;ll++)
			if (ll != icol) {
				dum=a[ll][icol];
				a[ll][icol]=0.0;
				for (l=1;l<=n;l++) a[ll][l] -= a[icol][l]*dum;
				for (l=1;l<=m;l++) b[ll][l] -= b[icol][l]*dum;
			}
	}
	for (l=n;l>=1;l--) {
		if (indxr[l] != indxc[l])
			for (k=1;k<=n;k++)
				SWAP(a[k][indxr[l]],a[k][indxc[l]]);
	}
	free_ivector(ipiv,1,n);
	free_ivector(indxr,1,n);
	free_ivector(indxc,1,n);
}

#undef SWAP

